from __future__ import annotations

import calendar
from dataclasses import dataclass
from datetime import datetime, date, timedelta, time
from typing import overload, Optional, Set


@dataclass(frozen=True)
class ContinuousTimeRange:
    start: datetime
    end: datetime

    def __post_init__(self):
        if self.start > self.end:
            raise ValueError("Start cannot be later than end.")

    def __contains__(self, moment: datetime) -> bool:
        return self.start <= moment <= self.end

    @property
    def length(self) -> timedelta:
        return self.end - self.start

    def __and__(self, other: ContinuousTimeRange) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(max(self.start, other.start), min(self.end, other.end))

    def __add__(self, other) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(min(self.start, other.start), max(self.end, other.end))

    def __ge__(self, other) -> bool:  # TODO TEST
        """
        Check if self (S) is a superset of other (O): S ⊇ O
        """
        return self.start <= other.start and other.end <= self.end

    def __le__(self, other) -> bool:  # TODO TEST
        """
        Check if self (S) is a subset of other (O): S ⊆ O
        """
        return other.start <= self.start and self.end <= other.end

    def intersects_with(self, other: ContinuousTimeRange) -> bool:
        return other.start in self or other.end in self or self <= other


class TimeRange:

    @overload
    def __init__(self):
        pass

    @overload
    def __init__(self, start: datetime, end: datetime):
        pass

    @overload
    def __init__(self, cls, start: datetime, duration: timedelta):
        pass

    def __init__(self, start: Optional[datetime] = None,
                 end: Optional[datetime] = None,
                 duration: Optional[timedelta] = None):
        self._periods: Set[ContinuousTimeRange] = set()
        if (start and not (end or duration)) or (end and duration):
            raise ValueError("Allowed combination of constructor parameters are:\n  "
                             "* Empty constructor\n  * `start` & `end`\n  * `start` & `duration`")
        if start:
            self._periods.add(ContinuousTimeRange(start, end or start + duration))

    def __eq__(self, other: TimeRange) -> bool:
        return self._periods == other._periods

    def __repr__(self) -> str:
        return ' + '.join(
            [f"TimeRange(start={repr(p.start)}, end={repr(p.end)})" for p in self._periods]
        ) or "TimeRange()"

    def __bool__(self) -> bool:
        """
        A TimeRange evaluates to True if its duration is grater than zero.
        """
        return len(self._periods) != 0

    def __contains__(self, moment: datetime) -> bool:
        return any([moment in p for p in self._periods])

    def __add__(self, other) -> TimeRange:
        """
        Find the union of the two TimeRange's: A ⋃ B.
        """
        if type(self) != type(other):
            return NotImplemented
        intersectionless_periods: Set[ContinuousTimeRange] = set()
        for p in {*self._periods, *other._periods}:
            for per in intersectionless_periods:
                if p.intersects_with(per):
                    intersectionless_periods.discard(per)
                    intersectionless_periods.add(p + per)
                    break
            else:
                intersectionless_periods.add(p)
        result = TimeRange()
        result._periods = intersectionless_periods
        return result

    # def __and__(self, other: TimeRange) -> Optional[TimeRange]:
    #     """
    #     Find the intersection of the two TimeRange's: A ⋂ B.
    #     """
    #     raise NotImplementedError()

    @property
    def length(self) -> timedelta:
        return sum([p.length for p in self._periods], start=timedelta())

    @property
    def start(self) -> datetime:
        """
        Return the earliest moment in the TimeRange.
        """
        return min({p.start for p in self._periods})

    @property
    def end(self) -> datetime:
        """
        Return the latest moment in the TimeRange.
        """
        return max({p.end for p in self._periods})


class CalendarMonth(TimeRange):
    """
    Represent a calendar month.
    """

    def __init__(self, year: int, month: int):
        start_date = date(year, month, 1)
        end_date = self._last_date_of_month(year, month)
        super().__init__(self._to_datetime(start_date), self._to_datetime(end_date))
        self.year = year
        self.month = month

    def __repr__(self):
        return f'CalendarMonth(year={self.year}, month={self.month})'

    @property
    def next(self) -> CalendarMonth:
        """Return an instance of next month."""
        first_day_in_next_month = self.end.date() + timedelta(days=1)
        return CalendarMonth(first_day_in_next_month.year, first_day_in_next_month.month)

    @property
    def prev(self) -> CalendarMonth:
        """Return an instance of previous month."""
        last_day_in_previous_month = self.start.date() - timedelta(days=1)
        return CalendarMonth(last_day_in_previous_month.year, last_day_in_previous_month.month)

    @staticmethod
    def _last_date_of_month(year: int, month: int) -> datetime.date:
        _, last_day = calendar.monthrange(year, month)
        return date(year, month, last_day)

    @staticmethod
    def _to_datetime(d: date) -> datetime:
        return datetime.combine(d, time.min)



import unittest
from datetime import datetime, timedelta
from unittest import TestCase

from timeset import ContinuousTimeRange, TimeRange, CalendarMonth

date = (2021, 5, 20)
start = datetime(*date, 12, 12)
end = datetime(*date, 14, 12)
string_repr = "TimeRange(start=datetime.datetime(2021, 5, 20, 12, 12), end=datetime.datetime(2021, 5, 20, 14, 12))"


class ContinuousTimeRangeTest(unittest.TestCase):

    def setUp(self) -> None:
        self.timerange = ContinuousTimeRange(start, end)

    def test_initialization(self):
        with self.assertRaisesRegex(ValueError, "Start cannot be later than end."):
            ContinuousTimeRange(end, start)

    def test_contains(self):
        self.assertTrue(datetime(*date, 13, 12) in self.timerange)
        # Assert that endpoints are included
        self.assertTrue(start in self.timerange)
        self.assertTrue(end in self.timerange)

    def test_does_not_contain(self):
        self.assertFalse(datetime(*date, 22, 12) in self.timerange)

    def test_as_timedelta(self):
        self.assertEqual(ContinuousTimeRange(start, start).length, timedelta(hours=0))
        self.assertEqual(self.timerange.length, timedelta(hours=2))

    def test_intersection(self):
        union1 = self.timerange & ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 22, 12))
        self.assertEqual(union1, ContinuousTimeRange(datetime(*date, 13, 12), end))

        union2 = self.timerange & ContinuousTimeRange(start, start)
        self.assertEqual(union2, ContinuousTimeRange(start, start))

    def test_union(self):
        union1 = self.timerange + ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 15, 12))
        self.assertEqual(union1, ContinuousTimeRange(start, datetime(*date, 15, 12)))

        union2 = self.timerange + ContinuousTimeRange(start, start)
        self.assertEqual(union2, ContinuousTimeRange(start, end))

    def test_intersects_with(self):
        # Partial intersection
        self.assertTrue(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 15, 12))))

        # Self contains other
        self.assertTrue(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 12, 19), datetime(*date, 13, 12))))

        # Other contains self
        self.assertTrue(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 11, 12), datetime(*date, 18, 12))))

    def test_does_not_intersect_with(self):
        self.assertFalse(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 18, 12), datetime(*date, 19, 12))))

    def test_equality(self):
        self.assertTrue(self.timerange == ContinuousTimeRange(self.timerange.start, self.timerange.end))

    def test_inequality(self):
        self.assertFalse(self.timerange != ContinuousTimeRange(self.timerange.start, self.timerange.end))
        self.assertFalse(self.timerange == ContinuousTimeRange(self.timerange.start, self.timerange.start))

    def test_superset(self):
        # Is superset
        self.assertTrue(self.timerange >= ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 13, 14)))
        # Is superset (one edge in common)
        self.assertTrue(self.timerange >= ContinuousTimeRange(datetime(*date, 12, 12), datetime(*date, 13, 14)))
        # Is not superset (actually self is subset)
        self.assertFalse(self.timerange >= ContinuousTimeRange(datetime(*date, 12, 12), datetime(*date, 22, 14)))
        # Is not superset (no intersection)
        self.assertFalse(self.timerange >= ContinuousTimeRange(datetime(*date, 18, 12), datetime(*date, 22, 14)))

    def test_subset(self):
        # Is subset
        self.assertTrue(self.timerange <= ContinuousTimeRange(datetime(*date, 10, 12), datetime(*date, 18, 14)))
        # Is subset (one edge in common)
        self.assertTrue(self.timerange <= ContinuousTimeRange(datetime(*date, 12, 12), datetime(*date, 18, 14)))
        # Is not subset (actually self is superset)
        self.assertFalse(self.timerange <= ContinuousTimeRange(datetime(*date, 12, 12), datetime(*date, 13, 14)))
        # Is not subset (no intersection)
        self.assertFalse(self.timerange <= ContinuousTimeRange(datetime(*date, 18, 12), datetime(*date, 22, 14)))


class TimeRangeInitializationTest(unittest.TestCase):
    def test_empty_initialization(self):
        t = TimeRange()
        self.assertEqual(str(t), "TimeRange()")

    def test_initialization_with_start_and_end(self):
        t = TimeRange(start=start, end=end)
        self.assertEqual(str(t), string_repr)

    def test_initialization_with_start_and_duration(self):
        t = TimeRange(start=start, duration=timedelta(hours=8))
        self.assertEqual(t.length, timedelta(hours=8))
        self.assertTrue(start in t)
        self.assertTrue(start + timedelta(hours=8) in t)

    def test_start_end_integrity(self):
        with self.assertRaises(ValueError):
            TimeRange(start=end, end=start)


class TimeRangeBooleanTest(unittest.TestCase):
    def test_bool_true(self):
        t = TimeRange(start=start, end=end)
        self.assertTrue(t)

    def test_bool_false(self):
        t = TimeRange()
        self.assertFalse(t)


class TimeRangeTimedeltaTest(unittest.TestCase):
    def test_timedelta_two_hours(self):
        t = TimeRange(start=start, end=end)
        self.assertEqual(t.length, timedelta(hours=2))

    def test_timedelta_zero(self):
        t = TimeRange()
        self.assertEqual(t.length, timedelta(hours=0))


class TimeRangeContainsTest(unittest.TestCase):
    def test_contains(self):
        t = TimeRange(start=start, end=end)
        self.assertTrue(datetime(2021, 5, 20, 12, 13) in t)

    def test_does_not_contain(self):
        t = TimeRange(start=start, end=end)
        self.assertFalse(datetime(2021, 5, 20, 16, 12) in t)


class TimeRangeAdditionTest(unittest.TestCase):
    def test_no_overlap(self):
        t1 = TimeRange(start=datetime(*date, 12), duration=timedelta(hours=2))
        t2 = TimeRange(start=datetime(*date, 16), duration=timedelta(hours=2))
        self.assertEqual((t1 + t2).length, timedelta(hours=4))

    def test_one_contains_another(self):
        t1 = TimeRange(start=datetime(*date, 12), duration=timedelta(hours=2))
        t2 = TimeRange(start=datetime(*date, 13), duration=timedelta(hours=1))
        self.assertEqual((t1 + t2).length, timedelta(hours=2))

    def test_overlap(self):
        t1 = TimeRange(start=datetime(*date, 12), duration=timedelta(hours=3))
        t2 = TimeRange(start=datetime(*date, 13), duration=timedelta(hours=3))
        self.assertEqual((t1 + t2).length, timedelta(hours=4))


class TimeRangeStartEndTest(TestCase):
    compound_timerange = TimeRange(start, datetime(*date, 13, 12)) + TimeRange(datetime(*date, 13, 22), end)

    def test_start(self):
        self.assertEqual(self.compound_timerange.start, start)

    def test_end(self):
        self.assertEqual(self.compound_timerange.end, end)


class CalendarMonthTest(TestCase):
    def test_init(self):
        month = CalendarMonth(year=2021, month=10)
        self.assertEqual(month.year, 2021)
        self.assertEqual(month.month, 10)
        self.assertTrue(datetime(2021, 9, 1) not in month)
        self.assertTrue(datetime(2021, 10, 10) in month)
        self.assertTrue(datetime(2021, 11, 1) not in month)

    def test_next(self):
        month = CalendarMonth(year=2021, month=11)
        self.assertEqual(month.next, CalendarMonth(year=2021, month=12))
        self.assertEqual(month.next.next, CalendarMonth(year=2022, month=1))

    def test_prev(self):
        month = CalendarMonth(year=2021, month=2)
        self.assertEqual(month.prev, CalendarMonth(year=2021, month=1))
        self.assertEqual(month.prev.prev, CalendarMonth(year=2020, month=12))


if __name__ == '__main__':
    unittest.main()
